@@ -4,10 +4,11 @@ use strict;
 use vars qw( $VERSION @EXPORT_OK );
 use Encode;
 use Exporter 'import';
-use Scalar::Util;
 use XML::LibXML;
 
-$VERSION = '1.8';
+use utf8;
+
+$VERSION = '1.9';
 @EXPORT_OK = qw( &to_xml &word_tag_wrap );
 
 =head1 NAME
@@ -295,7 +296,8 @@ sub to_xml {
 		return undef;
 	}
 
-	if( ref( $opts{'number_conversion'} ) ne 'CODE' ) {
+	if( defined $opts{'number_conversion'} 
+		&& ref( $opts{'number_conversion'} ) ne 'CODE' ) {
 		warn "number_conversion argument must be a subroutine ref";
 		$opts{'number_conversion'} = undef;
 	}
@@ -319,9 +321,10 @@ sub to_xml {
 
 	my( $in_p, $in_div ) = ( undef, undef );
 	while(<FILE>) {
-		chomp;
+		s/\R+$//g; # chomp, no matter the newline char
 		next if /^\s*$/;
-		s/^\s*//;
+		s/^\s*//; # but keep trailing spaces - they're significant!
+		_current_context( $_ );
 		
 		if( /^=BODY/ ) {
 			$inbody = 1;
@@ -332,13 +335,14 @@ sub to_xml {
 			next;
 		}
 		
-		if( /^(\w+):(.*)$/ ) {
+		if( /^(\w+)\s*:\s*(.*)$/ ) {
 			# Make the header template substitution.
-			warn "Warning: header line $_ in body section" if $inbody;
+			_make_warning( "We are in the BODY section but this looks like a header" )
+				if $inbody;
 			my( $key, $val ) = ( lc( $1 ), $2 );
-			$val =~ s/^\s+//;
+			$val =~ s/\s+$//;
 			if( $key eq 'main' ) {
-				warn "Illegal key $key; not substituting";
+				warn "You cannot use '$key' as a substitution key!";
 			} else {
 				$tmpl =~ s/__${key}__/$val/gi;
 			}
@@ -366,7 +370,7 @@ sub to_xml {
 		my $doc;
 		my $ok = eval{ $doc = $parser->parse_string( $tmpl ); };
 		unless( $ok ) {
-		   warn "Parsing of new XML doc failed: $@";
+		   warn "Parsing of the new XML doc failed: $@";
 		   return undef;
 		}
 		$tmpl = decode( $doc->encoding, $doc->serialize( $opts{'format'} ) );
@@ -376,9 +380,11 @@ sub to_xml {
 
 sub _process_line {
 	my( $line, $in_div, $in_p, %opts ) = @_;
-	chomp $line;
-	
-	# Look for paragraph and div markers.  
+	chomp $line;	
+	my $checkline = $line; # This should be well-formed by the end
+	my $clopts = { %opts, 'nowarn' => 1 };
+	# Look for paragraph and div markers, i.e. our tags that can span multiple lines
+	# and that should be disregarded in the checkline. 
 	my $sigils = $opts{'sigils'};
 	my( $divsig, $pgsig ) = ( $sigils->{'div'}, $sigils->{'p'} );
 	while( $line =~ /\Q$divsig\E(\d*)/g ) {	  
@@ -388,7 +394,7 @@ sub _process_line {
 		$pos -= length( $divno ) if $divno;
 		
 		if( $in_div ) {
-			warn "Nonsensical division number at end-division tag"
+			_make_warning( "Nonsensical division number at end-division tag; are your '$divsig' tags balanced?" )
 				if $divno;
 			substr( $line, $pos, 1, '</div>' );
 		} else {
@@ -397,20 +403,27 @@ sub _process_line {
 		}
 		$in_div = !$in_div;
 	}
+	$checkline =~ s/\Q$divsig\E//g;
 			
 	while( $line =~ /\Q$pgsig\E/g ) {
 		my $p_str = '<' . ( $in_p ? '/' : '' ) . 'p>';
 		substr( $line, pos( $line ) - 1, 1, $p_str );
 		$in_p = !$in_p;
 	}
+	$checkline =~ s/\Q$pgsig\E//g;
 	
 	# Add and delete tags.	Do this first so that we do not stomp later
 	# instances of the dash (e.g. in XML comments).
-	while( $line =~ m|([-+])(\(([^\)]+)\))?(.*?)\1|g ) {
+	my $add_del_re = qr/([-+])(\(([^\)]+)\))?(.*?)\1/;
+	while( $line =~ /$add_del_re/g ) {
 		my( $op, $attr, $word ) = ( $1, $3, $4 );
 		#  Calculate starting position.
 		my $pos = pos( $line ) - ( length( $word ) + 2 );
+		#  Also for the checkline.
+		$checkline =~ /$add_del_re/g;
+		my $cpos = pos( $checkline ) - ( length( $word ) + 2 );
 		$pos -= ( length( $attr ) + 2 ) if $attr;
+		$cpos -= ( length( $attr ) + 2 ) if $attr;
 		# Figure out what the attribute string, if any, should be.
 		my $attr_str;
 		if( $attr && $attr =~ /\=/ ) {
@@ -423,6 +436,7 @@ sub _process_line {
 			. ( $attr_str ? " $attr_str" : '' )
 			. ">$word</" . ( $op eq '+' ? 'add' : 'del' ) . '>';
 		substr( $line, $pos, pos( $line ) - $pos, $interp_str );
+		substr( $checkline, $cpos, pos( $checkline ) - $cpos, $interp_str );
 	}
 
 	# All the tags that are not very special cases.
@@ -435,12 +449,14 @@ sub _process_line {
 			$tag_open = $tag_close = $tag_sig;
 		}
 		$line =~ s|\Q$tag_open\E(.*?)\Q$tag_close\E|_open_tag( $tag, $1, \%opts ) . "</$tag>"|ge;
+		$checkline =~ s|\Q$tag_open\E(.*?)\Q$tag_close\E|_open_tag( $tag, $1, $clopts ) . "</$tag>"|ge;
 	} 
 
 	# Standalone tags that aren't special cases.  Currently only cb.
 	foreach my $tag ( qw( cb ) ) {
 		my $tag_sig = $sigils->{$tag};	
 		$line =~ s|\Q$tag_sig\E|"<$tag/>"|ge;
+		$checkline =~ s|\Q$tag_sig\E|"<$tag/>"|ge;
 	} 
 	
 	
@@ -457,6 +473,7 @@ sub _process_line {
 		$pb_close = $pb_sig;
 	}
 	$line =~ s|^\Q$pb_open\E(\d+(.)?)\Q$pb_close\E\s*$|<pb n=\"$1\"/>|;
+	$checkline =~ s|^\Q$pb_open\E(\d+(.)?)\Q$pb_close\E\s*$|<pb n=\"$1\"/>|;
 	
 	# XML comments.	 Convert ## text ## to <!-- text -->
 	my $com_sig = $sigils->{'comment'};
@@ -467,6 +484,15 @@ sub _process_line {
 		$com_open = $com_close = $com_sig;
 	}
 	$line =~ s|\Q$com_open\E(.*?)\Q$com_close\E|<!--$1-->|g;
+	$checkline =~ s|\Q$com_open\E(.*?)\Q$com_close\E|<!--$1-->|g;
+	
+	# At this point our check-line should be well-balance. Send a warning if not.
+	my $parser = XML::LibXML->new();
+	my $fragment;
+	my $ok = eval{ $fragment = $parser->parse_balanced_chunk( "<TAG>$checkline</TAG>" ); };
+	unless( $ok ) {
+		_make_warning( "Sigils are not properly nested." );
+	}
 
 	# Finally, every line with text outside an XML tag must have a line
 	# break.  Any lb tag should be inside a cb, p, or div tag.
@@ -525,7 +551,8 @@ sub _open_tag {
 				if( $ok ) {
 					$nv = &$numconvert( uc( $fragment->textContent() ) );
 				} else {
-					warn "Unbalanced chunk in number tag: $text";
+					_make_warning( "Unbalanced chunk in number tag: $text" )
+						unless $opts->{nowarn};
 				}
 				$opened_tag = sprintf( '<num value="%s">%s', $nv, $text ) 
 					if defined $nv;
@@ -533,7 +560,8 @@ sub _open_tag {
 		}
 	} elsif ( $tag eq 'hi' ) {
 		unless( $arg ) {
-			warn "Empty argument passed to $tag tag";
+			_make_warning( "What kind of highlighting is this?" )
+				unless $opts->{nowarn};
 			$arg = 'DEFAULT';
 		}
 		$arg =~ s/\s+/_/g;
@@ -545,6 +573,26 @@ sub _open_tag {
 	return $opened_tag;
 }
 
+sub _make_warning {
+	my $message = shift;
+	my $context = _current_context();
+	my $warning = "($.) $context\n\tPossible problem! $message";
+	warn $warning;
+}
+
+
+## Utility to keep track of where we are
+{
+	my $curr_line;
+	
+	sub _current_context {
+		if( @_ ) {
+			$curr_line = shift;
+		}
+		return $curr_line;	
+	}
+}
+
 =item B<word_tag_wrap>( $xml_string )
 
 Takes a string containing a TEI XML document, and returns that
@@ -579,13 +627,22 @@ sub word_tag_wrap {
 		die "Passed argument is neither string, Document, or Element";
 	}
 		
-	my @textnodes = $root->getElementsByTagName( 'text' );
-	my %paragraphs;  # Cope with the fact that text nodes can be recursive
-	foreach my $t ( @textnodes ) {
-		map { $paragraphs{Scalar::Util::refaddr( $_ )} = $_ } $t->getElementsByTagName( 'p' );
+	my @paragraphs;
+	foreach my $t ( $root->getElementsByTagName( 'text' ) ) {
+		# Get the paragraphs in this text node; if it is already the same as
+		# a paragraph in our list, skip it.
+		foreach my $p ( $t->getElementsByTagName( 'p' ) ) {
+			next if grep { $_->isSameNode( $p ) } @paragraphs;
+			push( @paragraphs, $p );
+		}
 	}
-	foreach my $p ( values %paragraphs ) {
+	foreach my $p ( @paragraphs ) {
 		my $new_p = _wrap_children( $p );
+		# Remove the final whitespace from the paragraphs
+		my $lc = $new_p->lastChild;
+		if( ref( $lc ) eq 'XML::LibXML::Text' && $lc->data =~ /^\s+$/ ) {
+			$new_p->removeChild( $lc );
+		}
 		$p->replaceNode( $new_p );
 	}
 	
@@ -655,6 +712,7 @@ sub _wrap_children {
 				$word_node->setNamespace( $docns );
 				$word_node->appendText( $_ );
 				$new_node->appendChild( $word_node );
+				$new_node->appendText(' ');
 				# ...and keep it open until we find a new word or a space
 				$open_word_node = $word_node;
 			}
@@ -690,6 +748,7 @@ sub _wrap_children {
 				$segment_node->setAttribute( 'type', 'word' );
 				$segment_node->appendChild( $c );
 				$new_node->appendChild( $segment_node );
+				$new_node->appendText(' ');
 				# Keep it open in case there is not a leading space on the next
 				# text node.
 				$open_word_node = $segment_node;
@@ -700,6 +759,8 @@ sub _wrap_children {
 	return $new_node;	 
 }
 
+1;
+
 =back
 
 =head1 BUGS / TODO
